import string
import contractions
import requests
import os
import regex
from datetime import datetime

def fix_answer(s):
        def remove_articles(text):
            return regex.sub(r'\b(a|an|the)\b', ' ', text)

        def white_space_fix(text):
            return ' '.join(text.split())

        def remove_punc(text):
            exclude = set(string.punctuation)
            return ''.join(ch for ch in text if ch not in exclude)

        def lower(text):
            return text.lower()

        return white_space_fix(remove_articles(remove_punc(lower(s))))

def normalize_answer(text, lower=True):
    if isinstance(text, list):
        result = []
        for ele in text:
            ele = str(ele)
            if lower:
                ele = ele.lower()
            translator = str.maketrans('', '', string.punctuation)
            ele = ele.translate(translator)
            result.append(fix_answer(' '.join(ele.split())))
        return result
    else:
        text = str(text)
        if lower:
            text = text.lower()
        translator = str.maketrans('', '', string.punctuation)
        text = text.translate(translator)
        return fix_answer(' '.join(text.split()))

def calculate_f1_score_with_precision(str1, str2):
    # Split the strings into sets of words
    str1 = fix_answer(contractions.fix(normalize_answer(str1)))
    str2 = fix_answer(contractions.fix(normalize_answer(str2)))
    words_str1 = set(str1.split())
    words_str2 = set(str2.split())

    # Calculate true positives, false positives, and false negatives
    tp = len(words_str1.intersection(words_str2))
    fp = len(words_str1 - words_str2)
    fn = len(words_str2 - words_str1)

    # Calculate precision and recall
    precision = tp / (tp + fp) if (tp + fp) > 0 else 0
    recall = tp / (tp + fn) if (tp + fn) > 0 else 0

    # Calculate F1 score
    f1_score = 2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0

    return f1_score, precision, recall

def remove_punctuation(text):
    def remove_articles(text):
        return regex.sub(r'\b(a|an|the)\b', ' ', text)

    def white_space_fix(text):
        return ' '.join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return ''.join(ch for ch in text if ch not in exclude)

    def lower(text):
        try:
            return text.lower()
        except:
            return ''

    return white_space_fix(remove_articles(remove_punc(lower(text))))

def download_link(file, url, name):
    # if not os.path.isfile(file):
    # print("Downloading {}...".format(name))
    try:
        response = requests.get(url, stream=True)

        if response.status_code == 200:
            with open(file, 'wb') as f:
                f.write(response.content)
                # print("Download {} complete.".format(name))
        else:
            print("Failed to download the model. Status code:", response.status_code)
    except:
        if not os.path.isfile(file):
            print("Failed to download the model. Check your internet connection.")


def file_needs_update(url, file_path):
    """
    Check if the file at the given path needs to be updated based on the
    Last-Modified header from the file URL.
    """
    try:
        response = requests.head(url)
        if response.status_code == 200 and 'Last-Modified' in response.headers:
            remote_last_modified = requests.utils.parsedate_to_datetime(response.headers['Last-Modified'])
            if not os.path.exists(file_path):
                return True  # File does not exist, needs download.
            local_last_modified = datetime.fromtimestamp(os.path.getmtime(file_path), tz=remote_last_modified.tzinfo)
            return remote_last_modified > local_last_modified
    except requests.RequestException as e:
        # print(f"Error checking if file needs update: {e}")
        pass
    return False  # Default to not updating if we can't determine.


from .f1 import f1_score_with_precision_recall
from .utils.tools import *
import joblib
from scipy.sparse import hstack
import numpy as np
import os
import contractions
import requests


class PEDANT:
    def __init__(self): 
        current_dir = os.path.dirname(__file__) 
        model_dir = os.path.join(current_dir, 'classifier') 
        model_path = os.path.join(model_dir, 'lr_classifier.pkl') 
        vectorizer_path = os.path.join(model_dir, 'tf-idf_vectorizer.pkl') 

        clf_url = 'https://github.com/zli12321/qa_metrics/raw/master/qa_metrics/classifier/lr_classifier' 
        vectorizer_url = 'https://github.com/zli12321/qa_metrics/raw/master/qa_metrics/classifier/tf-idf_vectorizer' 
        # if not os.path.exists(model_path) or file_needs_update(clf_url, model_path) == True: 
        if True:
            if not os.path.exists(model_dir): 
                os.makedirs(model_dir)
            try:
                # print('Downloaded model path: ', model_path) 
                download_link(model_path, clf_url, 'PANDA model') 
                download_link(vectorizer_path, vectorizer_url, 'PANDA evaluation model tokenizer') 
            except:
                pass
        

        self.model = joblib.load(model_path)
        self.tokenizer = joblib.load(vectorizer_path)

    def download_latest_model(self):
        current_dir = os.path.dirname(__file__) 
        model_dir = os.path.join(current_dir, 'classifier') 
        model_path = os.path.join(model_dir, 'lr_classifier.pkl') 
        vectorizer_path = os.path.join(model_dir, 'tf-idf_vectorizer.pkl') 

        if not os.path.exists(model_dir): 
            os.makedirs(model_dir) 

        clf_url = 'https://github.com/zli12321/qa_metrics/raw/master/qa_metrics/classifier/lr_classifier' 
        vectorizer_url = 'https://github.com/zli12321/qa_metrics/raw/master/qa_metrics/classifier/tf-idf_vectorizer'
        print('Downloaded model path: ', model_path) 
        download_link(model_path, clf_url, 'PANDA model') 
        download_link(vectorizer_path, vectorizer_url, 'PANDA evaluation model tokenizer') 
        self.model = joblib.load(model_path) 
        self.tokenizer = joblib.load(vectorizer_path)

    '''
    Return the confidence score between the reference and candidate answers. 
    reference, candidate, and question are strings.
    '''
    def get_score(self, reference, candidate, question):
        reference = remove_punctuation(normalize_answer(str(reference)))
        candidate = remove_punctuation(normalize_answer(str(candidate)))

        if reference in candidate:
            return 1.0

        input_texts = []
        f1_scores, precisions, recalls = [], [], []
        curr_input_text = "[CLS] " + candidate + " [SEP] " + reference + " [SEP] " + remove_punctuation(normalize_answer(str(question))) + " [SEP]"
        input_texts.append(curr_input_text)
        f, p, r = calculate_f1_score_with_precision(str(candidate), str(reference))
       
        f1_scores.append(f)
        precisions.append(p)
        recalls.append(r)

        f1_scores = np.array(f1_scores).reshape(-1, 1)
        precisions = np.array(precisions).reshape(-1, 1)
        recalls = np.array(recalls).reshape(-1, 1)

        texts = self.tokenizer.transform(input_texts)

        '''
        Concatenate text features with f1 features
        '''
        features = hstack([texts, f1_scores, precisions, recalls])
        pred_probas = self.model.predict_proba(features)

        return pred_probas[0][0]


    '''
    Returns the classifier confidence score for the candidate answer matching judgment if the reference and candidate answers
    are lists. The reference and candidate answers can lists of strings or just strings. The question is a string.
    '''
    def get_scores(self, reference, candidate, question):
        # Calculate the F1 score between the referee and candidate
        confidence_scores = {}
        if isinstance(reference, list) and isinstance(candidate, list):
            references = reference
            candidates = candidate

            
            for reference in references:
                for candidate in candidates:
                    score = self.get_score(reference, candidate, question)
                    confidence_scores[reference] = {}
                    confidence_scores[reference][candidate] = score
                            
            return confidence_scores

            # for candidate in candidates:
            #     input_texts = []
            #     f1_scores, precisions, recalls = [], [], []
            #     for reference in references:

            #         input_texts.append("[CLS] " + candidate + " [SEP] " + reference + " [SEP] " + question + " [SEP]")
            #         f1_results = f1_score_with_precision_recall(reference, candidate)
            #         f, p, r = f1_results['f1'], f1_results['precision'], f1_results['recall']
            #         f1_scores.append(f)
            #         precisions.append(p)
            #         recalls.append(r)

            #     f1_scores = np.array(f1_scores).reshape(-1, 1)
            #     precisions = np.array(precisions).reshape(-1, 1)
            #     recalls = np.array(recalls).reshape(-1, 1)

            #     texts = self.tokenizer.transform(input_texts)

            #     '''
            #     Concatenate text features with f1 features
            #     '''
            #     features = hstack([texts, f1_scores, precisions, recalls])
            #     pred_probas = self.model.predict_proba(features)

            #     for i in range(len(pred_probas)):
            #         if references[i] not in confidece_scores:
            #             confidece_scores[references[i]] = {}
            #         confidece_scores[references[i]][candidate] = pred_probas[i][0]
                            
            # return confidece_scores
        elif isinstance(reference, list):
            references = reference
            for reference in references:
                score = self.get_score(reference, candidate, question)
                confidence_scores[reference] = {}
                confidence_scores[reference][candidate] = score
                            
            return confidence_scores

            #     input_texts.append("[CLS] " + candidate + " [SEP] " + reference + " [SEP] " + question + " [SEP]")
            #     f1_results = f1_score_with_precision_recall(reference, candidate)
            #     f, p, r = f1_results['f1'], f1_results['precision'], f1_results['recall']
            #     f1_scores.append(f)
            #     precisions.append(p)
            #     recalls.append(r)

            # f1_scores = np.array(f1_scores).reshape(-1, 1)
            # precisions = np.array(precisions).reshape(-1, 1)
            # recalls = np.array(recalls).reshape(-1, 1)

            # texts = self.tokenizer.transform(input_texts)

            # '''
            # Concatenate text features with f1 features
            # '''
            # features = hstack([texts, f1_scores, precisions, recalls])
            # pred_probas = self.model.predict_proba(features)

            
            # for i in range(len(pred_probas)):
            #     confidece_scores[references[i]] = {}
            #     confidece_scores[references[i]][candidate] = pred_probas[i][0]
                            
            # return confidece_scores
        elif isinstance(candidate, list):
            candidates = candidate
            for candidate in candidates:
                score = self.get_score(reference, candidate, question)
                confidence_scores[reference] = {}
                confidence_scores[reference][candidate] = score
                            
            return confidence_scores
        else:
            confidence_scores[reference] = {}
            confidence_scores[reference][candidate] = self.get_score(reference, candidate, question)

            return confidence_scores

    '''
    Given a list of reference, candidate, and a question, return the pair with the highest confidence score.
    '''
    def get_highest_score(self, reference, candidate, question):
        confidence_scores = self.get_scores(reference, candidate, question)

        max_score = -1
        max_pair = (None, None)

        for reference, candidates in confidence_scores.items():
            for candidate, score in candidates.items():
                if score > max_score:
                    max_score = score
                    max_pair = (reference, candidate)

        return max_pair, max_score

    '''
    Input your reference and candidate answers, and the question.
    Return True if the candidate answer is deemed correct. Else, False.
    '''
    def evaluate(self, reference, candidate, question):
        # Calculate the F1 score between the referee and candidate
        judgment = False
        if isinstance(reference, list) and isinstance(candidate, list):
            references = [remove_punctuation(normalize_answer(str(ele))) for ele in reference]
            candidates = [remove_punctuation(normalize_answer(str(ele))) for ele in candidate]
            question = remove_punctuation(normalize_answer(str(question)))

            for candidate in candidates:
                if judgment == False:
                    input_texts = []
                    f1_scores, precisions, recalls = [], [], []
                    for reference in references:
                        if reference in candidate:
                            return True
                        input_texts.append("[CLS] " + candidate + " [SEP] " + reference + " [SEP] " + question + " [SEP]")
                        f1_results = f1_score_with_precision_recall(reference, candidate)
                        f, p, r = f1_results['f1'], f1_results['precision'], f1_results['recall']
                        f1_scores.append(f)
                        precisions.append(p)
                        recalls.append(r)

                        f1_scores = np.array(f1_scores).reshape(-1, 1)
                        precisions = np.array(precisions).reshape(-1, 1)
                        recalls = np.array(recalls).reshape(-1, 1)

                        texts = self.tokenizer.transform(input_texts)

                    '''
                    Concatenate text features with f1 features
                    '''
                    features = hstack([texts, f1_scores, precisions, recalls])
                    preds = self.model.predict(features)

                    if "correct" in preds:
                        judgment = True
                            
            return judgment
        elif isinstance(reference, list):
            references = [remove_punctuation(normalize_answer(str(ele))) for ele in reference]
            candidate = remove_punctuation(normalize_answer(str(candidate)))
            input_texts, tagged_texts = [], []
            f1_scores, precisions, recalls = [], [], []
            for reference in references:
                if reference in candidate:
                    return True
                curr_input_text = "[CLS] " + candidate + " [SEP] " + reference + " [SEP] " + contractions.fix(normalize_answer(str(question))) + " [SEP]"
                input_texts.append(curr_input_text)
                f, p, r = calculate_f1_score_with_precision(str(candidate), str(reference))
                f1_scores.append(f)
                precisions.append(p)
                recalls.append(r)

            f1_scores = np.array(f1_scores).reshape(-1, 1)
            precisions = np.array(precisions).reshape(-1, 1)
            recalls = np.array(recalls).reshape(-1, 1)

            texts = self.tokenizer.transform(input_texts)

            result = hstack([texts, f1_scores, precisions, recalls])
            output = self.model.predict(result)


            return "correct" in output
        elif isinstance(candidate, list):
            candidates = [remove_punctuation(normalize_answer(str(ele))) for ele in candidate]
            reference = remove_punctuation(normalize_answer(str(reference)))
            question = remove_punctuation(normalize_answer(str(question)))

            input_texts = []
            f1_scores, precisions, recalls = [], [], []
            for candidate in candidates:
                if reference in candidate:
                    return True
                input_texts.append("[CLS] " + candidate + " [SEP] " + reference + " [SEP] " + question + " [SEP]")
                f1_results = f1_score_with_precision_recall(reference, candidate)
                f, p, r = f1_results['f1'], f1_results['precision'], f1_results['recall']
                f1_scores.append(f)
                precisions.append(p)
                recalls.append(r)

            f1_scores = np.array(f1_scores).reshape(-1, 1)
            precisions = np.array(precisions).reshape(-1, 1)
            recalls = np.array(recalls).reshape(-1, 1)

            texts = self.tokenizer.transform(input_texts)

            '''
            Concatenate text features with f1 features
            '''
            features = hstack([texts, f1_scores, precisions, recalls])
            preds = self.model.predict(features)

            if "correct" in preds:
                judgment = True
                            
            return judgment
        else:
            reference = remove_punctuation(normalize_answer(str(reference)))
            candidate = remove_punctuation(normalize_answer(str(candidate)))
            question = remove_punctuation(normalize_answer(str(question)))

            if reference in candidate:
                return True

            input_texts = []
            f1_scores, precisions, recalls = [], [], []
            input_texts.append("[CLS] " + candidate + " [SEP] " + reference + " [SEP] " + question + " [SEP]")
            f1_results = f1_score_with_precision_recall(reference, candidate)
            f, p, r = f1_results['f1'], f1_results['precision'], f1_results['recall']
            f1_scores.append(f)
            precisions.append(p)
            recalls.append(r)

            f1_scores = np.array(f1_scores).reshape(-1, 1)
            precisions = np.array(precisions).reshape(-1, 1)
            recalls = np.array(recalls).reshape(-1, 1)

            texts = self.tokenizer.transform(input_texts)

            '''
            Concatenate text features with f1 features
            '''
            features = hstack([texts, f1_scores, precisions, recalls])
            preds = self.model.predict(features)

            if "correct" in preds:
                judgment = True
                            
            return judgment